@@ -38,5 +38,6 @@ except ImportError:
     from lmdb.cffi import __doc__
 
 del os
-__all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error']
-__version__ = '0.61'
+__all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error',
+           'enable_drop_gil']
+__version__ = '0.62'
