@@ -34,7 +34,8 @@ import weakref
 
 import cffi
 
-__all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error']
+__all__ = ['Environment', 'Cursor', 'Transaction', 'open', 'Error',
+           'enable_drop_gil']
 
 _ffi = cffi.FFI()
 _ffi.cdef('''
@@ -302,6 +303,20 @@ def _mvstr(mv):
     """Convert a MDB_val cdata to Python bytes."""
     return _ffi.buffer(mv.mv_data, mv.mv_size)[:]
 
+def enable_drop_gil():
+    """
+    Arrange for the global interpreter lock to be released during database IO.
+    This flag is ignored and always assumed to be ``True`` on cffi. Note this
+    can only be set once per process.
+
+    Continually dropping and reacquiring the GIL may incur unnecessary overhead
+    in single-threaded programs. Since Python intra-process concurrency is
+    already limited, and LMDB supports inter-process access, programs using
+    LMDB will achieve better throughput by forking rather than using threads.
+
+    *Caution:* this function should be invoked before any threads are created.
+    """
+
 
 class Environment(object):
     """
@@ -373,28 +388,34 @@ class Environment(object):
             Maximum number of databases available. If 0, assume environment
             will be used as a single database.
 
-        `initial_readers`:
-            Indicates the number of read transaction to preallocate in the
-            `reset` state. When `max_readers` is large and there are many open
-            read transactions, starting a read will incur a penalty due to the
-            linear scan required on the readers table. In this case read
-            transactions may be accelerated through preallocation.
+        `max_spare_txns`:
+            Read-only transactions to cache after becoming unused. Caching
+            transactions avoids two allocations, one lock and linear scan
+            of the shared environment per invocation of :py:meth:`begin`,
+            :py:class:`Transaction`, :py:meth:`get`, :py:meth:`gets`, or
+            :py:meth:`cursor`. Should match the process's maximum expected
+            concurrent transactions (e.g. thread count).
 
-            Currently unused in both CPython and CFFI.
+            *Note:* ignored on cffi.
 
-        `spare_readers`:
-            Indicates the number of read transactions to maintain in the
-            `reset` state. When a read transaction completes and the internal
-            `spare_readers` list is smaller than this number, the transaction
-            is reset and appended to the list. Otherwise, the transaction is
-            aborted and its associated reader slot is freed.
+        `max_spare_cursors`:
+            Read-only cursors to cache after becoming unused. Caching cursors
+            avoids two allocations per :py:class:`Cursor` or :py:meth:`cursor`
+            or :py:meth:`Transaction.cursor` invocation.
 
-            Currently unused in both CPython and CFFI.
+            *Note:* ignored on cffi.
+
+        `max_spare_iters`:
+            Iterators to cache after becoming unused. Caching iterators avoids
+            one allocation per :py:class:`Cursor` ``iter*`` method invocation.
+
+            *Note:* ignored on cffi.
     """
     def __init__(self, path, map_size=10485760, subdir=True,
             readonly=False, metasync=True, sync=True, map_async=False,
             mode=0o644, create=True, writemap=False, max_readers=126,
-            max_dbs=0, initial_readers=0, spare_readers=0):
+            max_dbs=0, max_spare_txns=1, max_spare_cursors=32,
+            max_spare_iters=32):
         envpp = _ffi.new('MDB_env **')
 
         rc = mdb_env_create(envpp)
